/*
disk_io.c
HSMCI driver for SD cards for SAM3 

Written by Robert Wood
Jan 2013 */ 
#include "disk_io.h"

#define ROB_INIT

unsigned long RelativeCardNumberValue;
unsigned long RelativeCardNumberValueShiftedLeft;
unsigned char CardType;
unsigned long NumberOfSectorsOnCard;

static DSTATUS Stat = STA_NOINIT;	/* Disk status */

//===================================================================================================
//							INITIALIZATION/STATUS FUNCS
//===================================================================================================


void InitHSCMI(void)
{

// Enable the clock for the HSCMI peripheral  
    PMC->PMC_PCER0 = 1<<ID_HSMCI;	

 // First set the pins to be connected to the HSCMI peripheral
    SetGPIOPeripheral(PIOA, 26, PERIPHERAL_C);
    SetGPIOPeripheral(PIOA, 27, PERIPHERAL_C);
    SetGPIOPeripheral(PIOA, 28, PERIPHERAL_C);
    SetGPIOPeripheral(PIOA, 29, PERIPHERAL_C);
    SetGPIOPeripheral(PIOA, 30, PERIPHERAL_C);
    SetGPIOPeripheral(PIOA, 31, PERIPHERAL_C);    
// And set GPIOA pin 6 to be a GPIO input as a card detect pin    
    SetGPIOFunction(PIOA, 6, INPUT);

    REG_HSMCI_CR = HSMCI_CR_SWRST;    // Software reset of the HSCMI peripheral
    REG_HSMCI_WPMR = 0x4D434900;  // A key to enable certain registers to be altered	
    HSMCI->HSMCI_CR = HSMCI_CR_MCIEN; // //enable MCI, disable power saving mode

    
    HSMCI->HSMCI_DTOR = 0x0F | HSMCI_DTOR_DTOMUL_1048576; // Set for maximum timeout for block transfers

// Attempt to set initial speed to 400kHz and specify the block size. Block size is set in header file and set to 512bytes
    REG_HSMCI_MR = 79 | HSMCI_MR_RDPROOF | HSMCI_MR_WRPROOF | (uint32_t)(BLOCK_SIZE << 16);       
// Set the peripheral up to transfer data in four bit mode and use slot A (there is only one slot on the SAM3S8)
    HSMCI->HSMCI_SDCR = HSMCI_SDCR_SDCSEL_SLOTA | HSMCI_SDCR_SDCBUS_4;		
    
        
    REG_HSMCI_CR = 0x01;    // Enable the HSMCI
   
  
}

/**********************************************************************************************
* Returns whether card is HD or SD (HD = high density, 4GB or higher)                         *
• A Global Variable, CardType is set to be HD_CARD or SD_CARD                                 *

**********************************************************************************************/
unsigned long GetCardType(void)
{

    unsigned long Response;
    unsigned char i;
    for(i=0;i<100;i++)
        {
        delay_ms(10);                        
// Set the bus width of the card to be four bits wide. It should be noted that the card interface was set to be 
// four bits wide previously, but that is only for data, not command responses. It's an extended command 
    if (SendExtendedCommand(ACMD41, 0, (0x00FF8000 | (1<<30)),&Response) == FAIL)
          return FAIL; 

    if(Response & (1<<31))	//if ready bit is set
        {
        //set SDHC flag when ready and CSS bits are set
        if(Response & (1<<30)) 
            CardType = HD_CARD; 
        else 
            CardType = SD_CARD; 
        return PASS;
        }    
    }
    
    return(FAIL);
}


//delay for less than 1ms
static void shortDelay(void)
{
	delay_us(250);
}


/**********************************************************************************************
* Read a single sector of the SD card. Pass the sector number and a pointer to the buffer     *
* the read data is to be stored in                                                            *
* This could be achieved with DMA of course, but in this case is not                          *
***********************************************************************************************/
unsigned long ReadSDSector(unsigned char *Buffer, unsigned long lba)
{

    unsigned long Response; // Unused here, but SendCommand needs it
    unsigned long SectorAddress;
    unsigned long wordsToRead;
    unsigned long temp;
    unsigned char i;

    wordsToRead = BLOCK_SIZE / 4;	// Data is read back in words, so divide number of bytes by four
    HSMCI->HSMCI_BLKR = (unsigned long)(BLOCK_SIZE) << 16; // Ensure the block size 

// If it is a standard density card (2GB or less), then it is the actually address of the card (which must be on 
// a sector boundary presumably) that is passed, so we multiple the sector number by the sector size
    if (CardType == SD_CARD)  
        lba *= BLOCK_SIZE;

    SendCommand(CMD17,lba,&Response);  // CMD17 is read single block
      
    while(wordsToRead--)
        {
        while((HSMCI->HSMCI_SR & HSMCI_SR_RXRDY)==0);   // Wait for data to be received
        temp = HSMCI->HSMCI_RDR;          // Quickly store it before it gets overwritten
        for(i=0;i<4;i++)
            {
            *Buffer++ = temp;   // Store the four bytes of the word
            temp = temp>>8;            
            }
        }
    return PASS;

}


static uint32_t sdWriteBlock(unsigned char *Buffer, uint32_t lba)
{

    unsigned long Response;  // Not used, needed to call SendCommand
    unsigned long byteAddress; 
    unsigned long wordsToWrite;
    unsigned long WordToSend;
    unsigned char i;
    unsigned char *TempPtr;

    HSMCI->HSMCI_MR &= ~HSMCI_MR_PDCMODE;//disable PDC mode (if enabled)
    

    if (CardType == SD_CARD)
        lba  *= BLOCK_SIZE;    
    SendCommand(CMD24,lba,&Response);   
    
    wordsToWrite = BLOCK_SIZE / 4;

    while (wordsToWrite--)
        {
        while(!(HSMCI->HSMCI_SR & HSMCI_SR_TXRDY)); 
        TempPtr = (unsigned char*)&WordToSend;
        for(i=0;i<4;i++)              
            *TempPtr++ = *Buffer++;              
        HSMCI->HSMCI_TDR = WordToSend;
        }
    
    while((HSMCI->HSMCI_SR & HSMCI_SR_NOTBUSY) == 0); //wait for NOTBUSY bit set
    return PASS;

}


/**********************************************************************************************
* Each command has a fixed response length (sent on the biderectional MCCDA line). This       *
* routine is just a look-up table to return the response length. There are special cases      *
* for setting extra flags in the calling function, but there are three lengths:               *
* No reponse                                                                                  *
* 48 bits (one word)                                                                          *
* 136 bits (four words)                                                                       *
***********************************************************************************************/
unsigned long GetReponseLength(unsigned long Command)
{

    unsigned long RetVal;

    switch (Command)
        {
        case CMD0: RetVal = ZERO_LENGTH_RESPONSE; break;
        case CMD1: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE; break;
        case CMD2: RetVal = ONE_HUNDRED_AND_THIRTY_SIX_BIT_LENGTH_RESPONSE; break;
        case CMD3: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE; break;
        case CMD5: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE; break;
        case CMD7: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE; break;
        case CMD8: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE; break;
        case CMD9: RetVal = ONE_HUNDRED_AND_THIRTY_SIX_BIT_LENGTH_RESPONSE; break;
        case CMD13: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE; break;
        case CMD17: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE_FOR_DATA_READ_TRANSFER; break;
        case CMD24: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE_FOR_DATA_WRITE_TRANSFER; break;
        case CMD55: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE; break;
        case CMD58: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE; break;
        case HSMCI_CMDR_SPCMD_INIT: RetVal =  ZERO_LENGTH_RESPONSE; break;
        case ACMD6: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE; break;
        case ACMD41: RetVal = FORTY_EIGHT_BIT_LENGTH_RESPONSE; break;
        default: ERROR_INVALID_COMMAND; break;

        }

    return RetVal;

}


/**********************************************************************************************
* Send a command out over the card interface peripheral. It should be noted that all          *
* communication is done synchronously over the MCCDA line using the MCCK line. The data lines *
* are totally unused. The MCCDA line is bidirectional. Results come back in the HMSCI_RSPRx   *
* registers.                                                                                  *
* Results can be 48 or 136 bits long. A 48 bit message is just stored in HMSCI_RSPR0,         *
* whereas 136 is stored in HMSCI_RSPR0-3, but read out by reading HMSCI_RSPR0 four times.     *
* The length of a message is fixed, so looked up in the GetReponseLength routine.             *
* It should be noted that the CRC is taken care of by the peripheral. The user does not have  *
* to worry him or herself about it as long as the CRC error bit is not set!                   *
**********************************************************************************************/
unsigned char SendCommand(unsigned long Command, unsigned long Argument, unsigned long *Response)
{

    unsigned long ResponseLength;
    unsigned long status;


    ResponseLength = GetReponseLength(Command);

// Once the message length has been found, we need to set a bit pattern so the peripheral knows how many 
// response bits will be returned
    if (ResponseLength == ERROR_INVALID_COMMAND)
        return ERROR_INVALID_COMMAND;
    else if (ResponseLength == FORTY_EIGHT_BIT_LENGTH_RESPONSE)
        Command |= HSMCI_CMDR_RSPTYP_48_BIT;
    else if (ResponseLength == ONE_HUNDRED_AND_THIRTY_SIX_BIT_LENGTH_RESPONSE)
        Command |= HSMCI_CMDR_RSPTYP_136_BIT;
    else if (ResponseLength == FORTY_EIGHT_BIT_LENGTH_RESPONSE_FOR_DATA_READ_TRANSFER)
        Command |= (HSMCI_CMDR_RSPTYP_48_BIT | HSMCI_CMDR_TRCMD_START_DATA | HSMCI_CMDR_TRDIR);
    else if (ResponseLength == FORTY_EIGHT_BIT_LENGTH_RESPONSE_FOR_DATA_WRITE_TRANSFER)
        Command |= (HSMCI_CMDR_RSPTYP_48_BIT | HSMCI_CMDR_TRCMD_START_DATA);

    Command |= HSMCI_CMDR_MAXLAT;  // Add maximum latency, to give card longest possible time to respond

// Now all the parameters are known and registers ready, just write the Argument first and then the command. 
// Once the command is written the peripheral sends the message
    HSMCI->HSMCI_ARGR = Argument;
    HSMCI->HSMCI_CMDR = Command;

// Now wait for the response. The HSMCI_SR_CMDRDY bit in the HSMCI->HSMCI_SR register is set when the response
  // has been received. 
    do
	{
        status = HSMCI->HSMCI_SR;
	} while((status & HSMCI_SR_CMDRDY)==0);

 // There is a section of the status register than contains various error bits. It might be necessary/a good idea
 // to return the actual error?
      if(status & MCI_ERRORS_MASK)
          return FAIL;
      else  // So, we have a valid response, it needs to be read. 
         {
         if (ResponseLength == FORTY_EIGHT_BIT_LENGTH_RESPONSE)
            {
            *Response = HSMCI->HSMCI_RSPR[0];          
            }
         else if (ResponseLength == ONE_HUNDRED_AND_THIRTY_SIX_BIT_LENGTH_RESPONSE)
            {
            *Response++ = HSMCI->HSMCI_RSPR[0];            
            *Response++ = HSMCI->HSMCI_RSPR[0];                    
            *Response++ = HSMCI->HSMCI_RSPR[0];            
            *Response = HSMCI->HSMCI_RSPR[0];          
            }       
         return PASS;
         }  
}


/**********************************************************************************************
* The RCA is the relative card address which is given out on-the-fly by the card to the host. *
* The calling routine sends the pointer to the response register which picks out the data     *
* and stores it in two global register.                                                       *
**********************************************************************************************/
void GetRCA(unsigned long *Data)
{

    unsigned long D1;
    
    D1 = *Data;    

    RelativeCardNumberValue = ((D1 >> 16) & 0xFFFF);
    RelativeCardNumberValueShiftedLeft = D1 & 0xFFFF0000;
    
}

/**********************************************************************************************
* Once the peripheral has been initalised, the SD card then be initialised. This involves a   *
* number of arcane steps that are lised in the SD card specification document.                *
* This routine continually calls the SendCommand routine and checks the command has been      *
* successfully sent. If any command fails, this routine fails and returns FAIL. If all goes   *
* well it returns OK at the end of the routine.                                               *
**********************************************************************************************/
unsigned char InitiliaseCard(void)
{

    unsigned long Response[4];
    unsigned char i;
    unsigned long CommandRetries;
    unsigned long Error;
    unsigned long TestAddr;
    unsigned long BlockLength;
    
    SendCommand(HSMCI_CMDR_SPCMD_INIT, ARG_DONT_CARE,&Response[0]); //initialization command (74 clock cycles for initialization sequence)
    SendCommand(CMD0, ARG_DONT_CARE,&Response[0]);  // Reset the card
    if (SendCommand(CMD8,0x1AA,&Response[0]) == FAIL) // Set 2.7-3.3V operation, should also return 0x1AA        
        return FAIL;
    else
        {
        if ((Response[0] & 0x1FF) != 0x1AA)
            return FAIL;
        }

// Find out whether the card is SD or HD and set the relevant variable
   if (GetCardType() == FAIL)
        return(FAIL);

    delay_ms(30000);

    if (SendCommand(CMD2, ARG_DONT_CARE,&Response[0]) == FAIL)  // Ask for the CID, why and how to decode it is anyone's guess    
        return FAIL;

    if (SendCommand(CMD3, ARG_DONT_CARE,&Response[0]) == FAIL)  // Ask for the relative address, it is returned in Response[0]
        return FAIL;

    GetRCA(&Response[0]);   // Get the relative card address and store it in the RCA global variable
    delay_ms(3000);
    
    if (SendCommand(CMD9, RelativeCardNumberValueShiftedLeft,&Response[0]) == FAIL)  // Ask for the CSD register and glean card size etc
        return FAIL;
    else
        {
        BlockLength = ((Response[1] >> 16) & 0x0F);
        if (BlockLength==9)
            BlockLength=512;
        else if (BlockLength==10)
            BlockLength=1024;
        else if (BlockLength==11)
            BlockLength=2048;
         if ((Response[0] & 0xC0) == 0x40) // This is CSD_Struct which is 0 for SD v1/SD cards and 10 for v2/HD cards
            {	/// SDC ver 2.0            
            NumberOfSectorsOnCard = ((Response[1] << 16) & 0x3F0000); // Get the upper six bits from the CSR of the C_SIZE field
            NumberOfSectorsOnCard |= ((Response[2] >> 16) & 0xFFFF);
            //cs = csd[9] + ((WORD)csd[8] << 8) + ((DWORD)(csd[7] & 63) << 8) + 1;
            NumberOfSectorsOnCard *= BlockLength;
            }
        else
            {
            NumberOfSectorsOnCard = ((Response[1] << 2) & 0xFFC);
            NumberOfSectorsOnCard |= ((Response[2] >> 30) & 0x03);
            NumberOfSectorsOnCard *= BlockLength;
            }            
        }
    

//Now the card can be put into a higher speed more. It must be 400kHz or less for initialisation   
// MASTER_CLOCK_SPEED is set in defines.h as it can be used in several .c files 
// SD_BUS_SPEED is set in sd_hsmci.h
    HSMCI->HSMCI_MR &= ~HSMCI_MR_CLKDIV_Msk; //zero CLKDIV field
    HSMCI->HSMCI_MR |= GetHSMCIClockDivisionForRequiredSpeed(CHIP_FREQ_CPU_MAX,SD_BUS_SPEED);    

// Activate the card. Once the card has given its RCA out, it must be activated for data transfer
    if (SendCommand(CMD7, RelativeCardNumberValueShiftedLeft,&Response[0]) == FAIL)  
          return FAIL;

// Set the bus width of the card to be four bits wide. It should be noted that the card interface was set to be 
// four bits wide previously, but that is only for data, not command responses. It's an extended command 
    if (SendExtendedCommand(ACMD6, RelativeCardNumberValueShiftedLeft, 2, &Response[0]) == FAIL)
          return FAIL;         

    return PASS;
    
}

/**********************************************************************************************
* The data sheet defines the clock divisiion as follows:                                      *
• CLKDIV: Clock Divider                                                                       *
* High Speed MultiMedia Card Interface clock (MCCK or HSMCI_CK)                               *
* is Master Clock (MCK) divided by (2*(CLKDIV+1)).                                            *
**********************************************************************************************/
unsigned long GetHSMCIClockDivisionForRequiredSpeed(unsigned long ProcessorSpeed, unsigned long RequiredHSMCIClockSpeed)
{
      
    return (((ProcessorSpeed/(2*RequiredHSMCIClockSpeed)) - 1) & 0xFF);

}

/**********************************************************************************************
* Extended commands always send CMD55 first and then a "special" command.                     *
* Pass: the extended command; the number of the card (0 before initialisation and the         *
* relative address shifted left, after it; the argument for the extended command and the      *
* pointer to the response. As always a success or failure is returned                         *
**********************************************************************************************/
unsigned char SendExtendedCommand(unsigned long ExtendedCommand, unsigned long CardNumber, unsigned long Argument, unsigned long *Response)
{

    if (SendCommand(CMD55, CardNumber,Response) == FAIL)  
          return FAIL;
    if (SendCommand(ExtendedCommand, Argument,Response) == FAIL) 
          return FAIL;

    return PASS;

}

//===================================================================================================
//								PUBLIC/INTERFACE FUNCTIONS
//===================================================================================================


unsigned char InitialiseSD(void)
{
	
      InitHSCMI();    // First initialize MCI controller and IO pins
      if (InitiliaseCard() == PASS)   // then initaliase the card itself
          {
          Stat = 0;
          return PASS;
          }
      else
          return FAIL;

}



uint8_t sdWriteBlocks(void *buffer, uint32_t lba, uint32_t blocksToWrite)
{

      uint8_t *pBuf = (uint8_t*)buffer;
      while(blocksToWrite--)
          {
          if(sdWriteBlock(pBuf,lba) == FAIL) 		  		  
              return(FAIL);		  
          pBuf+=BLOCK_SIZE;
          lba++;
          }
    return PASS;
}

/**********************************************************************
* Read a number of sectors of the SD card. It's really just a       	*
* wrapper for ReadSDSector which reads a single sector			*
***********************************************************************/
unsigned char ReadSDCardSectors(unsigned char *buffer, unsigned long lba,unsigned long NumberOfSectorsToRead)
{

    while(NumberOfSectorsToRead--)
        {                
        if(ReadSDSector(buffer,lba) == FAIL)
            return FAIL;
        buffer+=BLOCK_SIZE;
        lba++;
        }
}


/**********************************************************************
* Nasty little delay routines, should be changed to use timers.	*
***********************************************************************/
void delay_us(unsigned long DelayTime)
{

    unsigned long i;
        for (i=0;i<DelayTime;i++);

}

void delay_ms(unsigned long DelayTime)
{

    
    unsigned long i;
    unsigned long j;


    
    for (j=0;j<3;j++)
        {
        for (i=0;i<DelayTime;i++);
        }

}
/**********************************************************************
* Write to the disk. A wrapper to knit Atmel code to Chan code      	*
* BYTE drv,			 Physical drive nmuber (0) 	    	*
	const BYTE *buff,	 Pointer to the data to be written 	*
	DWORD sector,		 Start sector number (LBA) 		*
	BYTE count			 Sector count (1..128) 		*
********************************************************************/
DRESULT disk_write (BYTE drv, BYTE *buff,	DWORD sector, BYTE count)
{

     sdWriteBlocks(buff, sector, count);


}

/********************************************************************
* disk_read is just a wrapper because the FAT file libraries used   *
* call this routine. This in turn directly calls the                *
* ReadSDCardSectors                                                 *
* The weird defines such as BYTE and DWORD are kept so it is not    *
* necessary to alter the existing FAT libraries at all               *
* BYTE drv,			Physical drive number (Always 0)    *
* BYTE *buff,	Pointer to the data buffer to store read data       *
* DWORD sector, First sector number                                 *
* BYTE count	 Number of sectors to read, will ususally be 1      *
********************************************************************/
DRESULT disk_read(BYTE drv, BYTE *buff, DWORD sector, BYTE count)
{

    ReadSDCardSectors(buff,sector,count);

}


/************************************************************************
* Get Disk Status                                                       *
* drv = drive number (always 0)                                         *
* Keeping the odd variable types such as BYTE so as not to necessitate  *
* changing the fat file software                                        *
*************************************************************************/
DSTATUS disk_status(BYTE drv)
{
	DSTATUS s;
	BYTE d;
        unsigned long Response;


	if (drv) return STA_NOINIT;

	/* Check if the card is kept initialized */
	s = Stat;
	if (!(s & STA_NOINIT)) 
            {
            if (SendCommand(CMD13, RelativeCardNumberValueShiftedLeft, &Response))                        
                s = STA_NOINIT;
            
           //deselect();
            }
	Stat = s;

	return s;
}

// A wrapper function for when the existing FAT file code calls the card initialisation
DSTATUS disk_initialize(BYTE drv)
{

    InitialiseSD();

}


/*****************************************************************************/
/* Miscellaneous Functions                                            	*
* BYTE drv,		 Physical drive nmuber (0)   				*
*	BYTE ctrl,		* Control code 					*
*	void *buff		* Buffer to send/receive control data    	*
*       void is passed because the result size is variable (yuk)        	*
******************************************************************************/
DRESULT disk_ioctl (BYTE drv,BYTE ctrl,void *buff)
{
	DRESULT res;
	BYTE n, csd[16];
	DWORD cs;
        unsigned long Response[4];
        unsigned long BlockLength;
        unsigned long CSize;


	if (disk_status(drv) & STA_NOINIT) return RES_NOTRDY;	/* Check if card is in the socket */

	res = RES_ERROR;
	switch (ctrl) 
            {
            case CTRL_SYNC : res = RES_OK; break;	/* Make sure that no pending write process */        
            case GET_SECTOR_SIZE :	/* Get sector size in unit of byte (WORD) */
		*(WORD*)buff = 512;
		res = RES_OK;
		break;                                    

            case GET_SECTOR_COUNT :	/* Get number of sectors on the disk (DWORD) - was calculated in InitiliaseCard() */
                *(DWORD*)buff = NumberOfSectorsOnCard;                
                    res = RES_OK;                    
                break;

            case GET_BLOCK_SIZE :	/* Get erase block size in unit of sector (DWORD) */
                  *(DWORD*)buff = 128;
                  res = RES_OK;
                  break;

		default:
			res = RES_PARERR;
	
              }
	

	return res;
}
